import 'dart:collection';

import 'package:routex_core/routex_core.dart';
import 'package:routex_web/src/route_impl.dart';
import 'package:routex_web/src/routing_context_impl.dart';
import 'package:routex_web/src/routing_context_wrapper.dart';


class RouterImpl implements Router {
  final Set<RouteImpl> routes = SplayTreeSet(routeComparator);
  final AtomicInteger orderSequence = AtomicInteger();

  static Function routeComparator = (RouteImpl o1, RouteImpl o2) {
    var x = o1.order;
    var y = o2.order;
    var compare = (x < y) ? -1 : ((x == y) ? 0 : 1);
    if (compare == 0) {
      if (x == y) {
        return 0;
      }
      return 1;
    }
    return compare;
  };

  final Map<int, Handler<RoutingContext>> _errorHandlers = {};

  @override
  Router errorHandler(int statusCode, Object errorHandler) {
    Objects.requireNonNull(errorHandler);
    _errorHandlers[statusCode] = RoutexAction.createHandler(errorHandler);
    return this;
  }

  Handler<RoutingContext> getErrorHandlerByStatusCode(int statusCode) =>
      _errorHandlers[statusCode];

  @override
  Route route(String path) =>
      RouteImpl(this, orderSequence.getAndIncrement(), path: path);

  @override
  Route routeWithRegex(String regex) =>
      RouteImpl(this, orderSequence.getAndIncrement(), regex: regex);

  @override
  Route get(String path) => RouteImpl(this, orderSequence.getAndIncrement(),
      path: path, content: HttpMethod.get);

  @override
  Route post(String path) => RouteImpl(this, orderSequence.getAndIncrement(),
      path: path, content: HttpMethod.post);

  @override
  Route put(String path) => RouteImpl(this, orderSequence.getAndIncrement(),
      path: path, content: HttpMethod.put);

  @override
  Route delete(String path) => RouteImpl(this, orderSequence.getAndIncrement(),
      path: path, content: HttpMethod.delete);

  void add(RouteImpl route) {
    routes.add(route);
  }

  @override
  RoutingExecution handle(HttpServerRequest request, dynamic payload) =>
      RoutingExecution(RoutingContextImpl(null, this, request, routes));

  Iterator<RouteImpl> iterator() => routes.iterator;

  @override
  Router mountSubRouter(String mountPoint, Router subRouter) {
    if (mountPoint.endsWith('*')) {
      throw IllegalArgumentException(
          "Don't include * when mounting subrouter");
    }
    if (mountPoint.contains(':')) {
      throw IllegalArgumentException(
          "Can't use patterns in subrouter mounts");
    }
    route(mountPoint + '*')
        .handler(subRouter.handleContext)
        .failureHandler(subRouter.handleFailure);
    return this;
  }

  @override
  void handleFailure(RoutingContext ctx) => RoutingContextWrapper(
        getAndCheckRoutePath(ctx),
        ctx.request(),
        routes,
        ctx,
      ).next();

  @override
  void handleContext(RoutingContext ctx) => RoutingContextWrapper(
        getAndCheckRoutePath(ctx),
        ctx.request(),
        routes,
        ctx,
      ).next();

  String getAndCheckRoutePath(RoutingContext ctx) {
    var currentRoute = ctx.currentRoute();
    var path = currentRoute.getPath();
    if (path == null) {
      throw Exception(
          'Sub routers must be mounted on constant paths (no regex or patterns)');
    }
    return path;
  }
}
